package admin;

import edu.neu.ccs.demeterf.http.server.*;
import edu.neu.ccs.demeterf.http.classes.*;
import edu.neu.ccs.demeterf.lib.*;
import scg.game.Game;
import scg.game.Tournament;
import scg.game.Tournament.Player;
import scg.game.Game.PlayerStore;
import scg.gen.*;
import scg.*;
import reg.RegServer;
import java.io.*;
import logging.Logger;
import java.util.Date;
import java.util.Timer;
import java.util.TimerTask;

/** Administrator Status response Server */
@Server
public class AdminStatus extends HTML {

    /** The commmon Page Title for responses */
    private final String TITLE = " Administrator Status";
    /** The commmon Page Title for responses */
    private final String ResultsPrefix = "final_results";
    /** Port where Status will Listen */
    @Port
    private final int port;
    /** Maximum Status request size... should be small */
    @MaxMessageSize
    private final int maxMessage = 1024;
    /** When the competition will start */
    Date start;
    /** The current Registration State */
    private Option<RegServer> reg = Option.none();
    /** The current Game State (will be mutated) */
    private Option<Game> game = Option.none();
    /** The current Game State (will be mutated) */
    private Option<Tournament> tourn = Option.none();
    /** Remember to write out the status to a globally accessable Page */
    private Timer remember;
    
    /** Crate an AdminStatus Server */
    public AdminStatus(int p, Date st, boolean fileStatus) {
        port = p;
        start = st;
        remember = new Timer();
        if(fileStatus){
            remember.scheduleAtFixedRate(new TimerTask(){
                public void run(){
                    try{
                        String name = "current_status.html";
                        writeResults(logging.Logger.nullLog(), name+".tmp", true);
                        int res = Runtime.getRuntime().exec(new String[]{"/bin/cp",name+".tmp",name}).waitFor();
                        if(res != 0)
                            System.err.println("Unable to copy Status ("+res+")");
                    }catch(Exception e){
                        System.err.println("Error writing Status: "+e.getMessage());
                    }
                }
            }, 1000, 2000);
        }
    }

    /** Set the RegServer, so the Status knows the current Reg State */
    public synchronized void setReg(RegServer r){
        reg = Option.some(r);
    }

    /** Was Registration Finished? */
    private boolean regDone = false;

    /** After registration completes (or is skipped) */
    public synchronized void finishReg(){
        reg = Option.none();
        regDone = true;
    }

    /** Set the GameState */
    public synchronized void setGame(Game g){
        game = Option.some(g);
    }
    /** Set the TournState */
    public synchronized void setTourn(Tournament t){
        tourn = Option.some(t);
    }

    /** Was Registration Finished? */
    private boolean gameDone = false;

    /** After registration completes (or is skipped) */
    public synchronized void finishGame(){
        gameDone = true;
    }

    /** Get the Players/Stores */
    public List<PlayerStore> getPlayerStates(){
        synchronized (game) {
            if (!game.isSome())
                return List.<PlayerStore> create();
            return List.create(game.inner().getPlayersSortedByAccountOrKicked());
        }
    }
    /** Get the Players/Stores */
    public List<Player> getTournPlayers(){
        synchronized(tourn) {
            if (!tourn.isSome())return List.<Player> create();
            return List.create(tourn.inner().getResultSortedByPoints());
        }
    }

    /** Get the Game Config */
    public Config getConfig(){
        return game.isSome()?game.inner().getConfig():tourn.inner().getConfig();
    }
    /** Handle any Exceptions prior to dispatch */
    //@ExceptionHandler
    public HTTPResp exception(Exception e){
        Throwable t = Util.rootCause(e);
        return HTTPResp.error(HTTPResp.MIN_SERVER_ERROR, "EXCEPTION", "Request/Response Exception ["+
                t.getClass().getSimpleName()+"]");
    }
    
    /** Return an HTTPResp containing the current Game Status */
    @Path("/status")
    public synchronized HTTPResp statusReq(HTTPReq requ){
        String prefix = (gameDone ? "Final" : "Current");
        boolean refresh = requ.urlArgs().containsKey("refresh");
        try {
            return okHTMLResp(htmlRefreshPage((!refresh||gameDone)?0:5,prefix + TITLE, "\n div{ text-align:left;border:black 1px solid;padding:10px; }"
                    + "\n table,td{ border:blue 1px solid;padding:5px; }" + "\n td,th{ border-spacing:5px; }"
                    + "\n td.num{ text-align:center; }"
                    + "\n th{ font-style:italic;border:solid navy 1px; padding:5px;}" + "\n .red{ color:red; }"
                    + "\n .odd{ background-color:#D0D0D0; }" + "\n .current{ background-color:#40D040; }"
                    + "\n .kicked{ background-color:#E04040; }", wrap(wrap(prefix + TITLE, "h3")
                    + wrap("  Now: " + wrap(Util.printDate(Util.now()), "i"), "h4")
                    + "<br/>"
                    + wrap(boolStat("Registration", reg, regDone) + "\n" + boolStat("Game", game, gameDone) + "\n"
                            + "<br/><br/>" + statusRes(), "div"), "center")));
        } catch (Exception e) {
            return errHTMLResp("Exception", errorPage("<div style=\"text-align:left;\">" + e + "<br/>"
                    + java.util.Arrays.toString(e.getStackTrace()) + "</div>"));
        }
    }

    /** Return the body of the Admin Status */
    public synchronized String statusRes(){
        // Reg is currently happening
        if (reg.isSome())
            return regStatus();

        // Game Hasn't Started, but Reg is finished (or Not Run)
        if (!game.isSome() && !tourn.isSome()) {
            return nogameStatus();
        }
        // Actual Game Status
        if(game.isSome())
            synchronized (game.inner()) {
                return gameStatus(game.inner().getCurrentPlayerName(), gameDone);
            }
        synchronized (tourn.inner()) {
            return tournStatus(tourn.inner().currentPlayer1(), tourn.inner().currentPlayer2(), gameDone);
        }
    }

    /** */
    private String boolStat(String what, Option<?> o, boolean done){
        return (wrap(what, "b") + ": " + (done ? "Complete" : (o.isSome() ? "In Progress" : "Not Run")) + "<br/>");
    }

    /** Time Until Start */
    private String untilStart(){
        long timeLeft = Math.max(0, (start.getTime() - Util.now().getTime()) / 1000);
        return wrap("Until Start: "
                + wrap(Util.twoDigits(timeLeft / 3600) + ":" + Util.twoDigits((timeLeft / 60) % 60) + ":"
                        + Util.twoDigits(timeLeft % 60), "i", "red"), "h4")
                + "<br/>";
    }

    /**  */
    private String regStatus(){
        final List<PlayerSpec> players = reg.inner().getPlayers().reverse();
        String countDown = untilStart();
        if (players.isEmpty())
            return wrap(countDown + wrap("No Registered Teams", "h3") + "\n", "center");
        return (wrap(countDown + wrap("Registered Teams", "h3") + "\n"
                + wrap(players.fold(new List.Fold<PlayerSpec, String>() {

                    int count = 1;

                    public String fold(PlayerSpec p, String s){
                        return s
                                + "\n"
                                + wrap(wrap("" + count++, "td") + wrap(p.getName(), "td") + wrap(p.getAddress(), "td"),
                                        "tr", (count % 2 == 1) ? "even" : "odd");
                    }
                }, wrap(wrap("#", "th") + wrap("Name", "th") + wrap("Address", "th"), "tr")), "table"), "center"));
    }

    /** Simple Status, before the Game Starts */
    private String nogameStatus(){
        return wrap(untilStart() + wrap("Game Not Running Yet", "h3"), "center");
    }

    /** Game status == A Players Table */
    private String gameStatus(final String currName, final boolean fin){
        List<PlayerStore> stores = getPlayerStates();
        boolean anyKicked = stores.ormap(new List.Pred<PlayerStore>() {

            public boolean huh(PlayerStore e){
                return e.wasKicked();
            }
        });
        String timeLeft = fin ? "" : wrap("Turn Time Remaining: "
                + wrap(Math.max(0, (getConfig().getTurndur() - (System.currentTimeMillis() - game.inner()
                        .getTurnStart()) / 1000))
                        + " sec.", "i", "red"), "h4");
        String key = wrap((fin ? "" : wrap(wrap("&nbsp;&nbsp;", "td", "current") + wrap("Current Turn", "td"), "tr"))
                + (anyKicked ? wrap(wrap("&nbsp;&nbsp;", "td", "kicked") + wrap("Kicked Out!", "td"), "tr") : ""),
                "table");
        return ((anyKicked || !fin) ? key : "")
                + wrap(wrap("Game " + (fin ? "Complete" : "Round #" + game.inner().getCurrentRound()), "h3")
                        + wrap(timeLeft + (fin ? "Final" : "Current") + " Table", "h3")
                        + wrap(stores.fold(new List.Fold<PlayerStore, String>() {

                            int count = 1;

                            public String fold(PlayerStore str, String s){
                                PlayerSpec spec = str.getSpec();
                                int curr = count++;
                                return s
                                        + "\n"
                                        + wrap(wrap(str.wasKicked() ? "--" : "" + curr, "td", "num")
                                                + wrap(spec.getName(), "td")
                                                + wrap(Util.format(str.getAccount()), "td")
                                                + (fin ? "" : wrap("" + str.numOffered(), "td", "num")
                                                        + wrap("" + str.numAccepted(), "td", "num")
                                                        + wrap("" + str.numProvided(), "td", "num"))
                                                + wrap("&nbsp;" + str.getKickReason(), "td"), "tr",
                                                str.wasKicked() ? "kicked"
                                                        : spec.getName().equals(currName) ? "current"
                                                                : (count % 2 == 1) ? "even" : "odd");
                            }
                        }, wrap(wrap("Rank", "th")
                                + wrap("Name", "th")
                                + wrap("Account", "th")
                                + (fin ? "" : wrap("#Offered", "th") + wrap("#Accepted", "th")
                                        + wrap("#Provided", "th")) + wrap("Kickout Reason", "th"), "tr")), "table"),
                        "center");
    }

    /** Game status == A Players Table */
    private String tournStatus(final String play1, final String play2, final boolean fin){
        List<Player> players = getTournPlayers();
        return wrap(wrap("Tournament " + (fin ? "Complete" : ("Game #" + tourn.inner().getGameNum()
                + " of " + tourn.inner().numGames())), "h3")
                + wrap((fin ? "Final" : "Current") + " Table", "h3")
                + wrap(players.fold(new List.Fold<Player, String>() {

                    int count = 1;

                    public String fold(Player str, String s){
                        PlayerSpec spec = str.getProxy().getSpec();
                        int curr = count++;
                        return s
                                + "\n"
                                + wrap(wrap(""+curr, "td", "num")
                                        + wrap(spec.getName(), "td")
                                        + wrap(""+str.getResult(), "td", "num"), "tr",
                                        (!fin && (spec.getName().equals(play1) || spec.getName().equals(play2)))
                                        ? "current" : ((count % 2 == 1) ? "even" : "odd"));
                    }
                }, wrap(wrap("Rank", "th")
                        + wrap("Name", "th")
                        + wrap("Points", "th"), "tr")), "table"),
                "center");
}
    
    /** Write the HTML results file */
    public void writeResults(Logger log, boolean tourn){
        writeResults(log, ResultsPrefix
                + (tourn?"_tournament":"_group")+ Util.getFileNameSafeDate(Util.now())
                + ".html", false);
    }
    
    public void writeResults(Logger log, String file, boolean refresh){
        log.event("Writing HTML Results...");
        try {
            PrintStream out = new PrintStream(new FileOutputStream(file));
            out.print(this.statusReq(HTTPReq.Get("/status"+(refresh?"?refresh=true":""))).getBodyString());
            out.close();
            log.event("Done");
        } catch (Exception e) {
            log.event("Error Writing Results: " + e.getMessage());
        }
    }
}
