package reg;

import edu.neu.ccs.demeterf.http.classes.*;
import edu.neu.ccs.demeterf.http.server.*;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.net.Socket;
import java.util.Date;


import logging.Logger;
import scg.Util;
import scg.Constants;
import scg.gen.ParseException;
import scg.gen.PasswordFile;
import scg.gen.PlayersFile;
import scg.gen.PlayerSpec;
import edu.neu.ccs.demeterf.lib.List;
import edu.neu.ccs.demeterf.lib.Map;
import edu.neu.ccs.demeterf.lib.Option;
import edu.neu.ccs.demeterf.util.CLI;

/**
 * Player/Competition Registration Server. Waits for registration requests, and
 * responds accordingly.
 * <p>
 * The registration protocol is as follows:
 * <ul>
 * <li>The Player sends a {@link scg.gen.PlayerSpec PlayerSpec} as the body of
 * a POST to the RegServer, including a URL argument.</li>
 * <li>The full URL request will be: <quote>
 * <tt>RegServer.PATH_ENTRY+"?"+RegServer.PASS_URL_ARG+"="+PASSWORD</tt>
 *        </quote>
 *        where <tt>PASSWORD</tt> is the Player/Team password as registered with
 *        the Teaching Staff (to be described separately).</li>
 *    <li>If registration succeeds, an HTTP Response with code 200 (OK) will be
 *        sent back, with a short the Player's name as a body.<li>
 *  </ul>
 *  </p>
 */
@Server
public class RegServer extends scg.HTML {

    /** Listening (incoming) Port */
    @Port
    private int port;
    /** Max Message Size, 1K should do */
    @MaxMessageSize
    private int maxMessage = 1024;

    /** Password List */
    private final Map<String, String> passwds;
    /** Currently registered Players */
    private List<PlayerSpec> players = List.create();
    /** Logger for the Registration Process */
    private final Logger log;
    /** Where to save the Registration */
    private final String savefile;
    /** Where to save the Registration */
    private final int maxLogins;

    /**
     * Create a RegServer with the given Passwords. Players/Teams will submit a
     * Passsword, which we will 'encrypt' and store to verify the registration
     * for competitions.
     */
    private RegServer(int p, Map<String, String> ps, Logger l, String file, boolean load, int logins) {
        port = p;
        passwds = ps;
        log = l;
        savefile = file;
        maxLogins = logins;
        log.notify("Logins Allowed: "+logins);
        try {
            if (load)
                players = PlayersFile.parse(new FileInputStream(file)).getPlayers();
        } catch (Exception e) {
            players = List.create();
        }
    }

    /** Create a RegServer at the given Port. */
    public static RegServer create(Logger l, String savefile, boolean load, int logins) throws ParseException, IOException{
        return create(Constants.DEF_REG_PORT, l, savefile, load, logins);
    }

    /** Create a RegServer with the given Passwords at the Default Port. */
    public static RegServer create(Map<String, String> ps, Logger l, String savefile, boolean load, int logins){
        return create(Constants.DEF_REG_PORT, ps, l, savefile, load, logins);
    }

    /**
     * Create a RegServer at the given Port, with the given Passwords.
     * Players/Teams will submit a Passsword, which we will 'encrypt' and store
     * to verify the registration for competitions.
     */
    public static RegServer create(int p, Map<String, String> ps, Logger l, String savefile, boolean load, int logins){
        return new RegServer(p, ps, l, savefile, load, logins);
    }

    /**
     * Create a RegServer with the given Logger. Passwords will be loaded from
     * the usual File.
     */
    public static RegServer create(int port, Logger l, String savefile, boolean load, int logins)
            throws ParseException, IOException{
        PasswordFile file = PasswordFile.parse(new FileInputStream(Constants.PASS_FILE));
        return create(port, file.getPasswordMap(), l, savefile, load, logins);
    }

    /** Get the current list of registered Players */
    public synchronized List<PlayerSpec> getPlayers(){
        return players;
    }

    /** Handle any Exceptions prior to dispatch */
    @ExceptionHandler
    public HTTPResp exception(Exception e, Socket s){
        //Throwable t = Util.rootCause(e);
        //log.error("Request/Response Exception ["+t.getClass().getSimpleName()+"] from ["+s.getInetAddress()+"]");
        return null; //errTextResp("", "");
    }
    
    /** Main URL handler for the RegServer */
    @Path(Constants.REG_PATH_ENTRY)
    public synchronized HTTPResp registerResp(HTTPReq req, Socket sock){
        log.event("Registration Request from: " + sock.getInetAddress());
        try {
            PlayerSpec spec = PlayerSpec.parse(req.getBodyString());
            Map<String, String> args = req.urlArgs();
            // System.err.println(" PASS:
            // '"+args.get(Constants.PASS_URL_ARG)+"'");
            if (!passwds.containsKey(spec.getName())) {
                return errTextResp("Unknown Team", log.error("Unknown Team Name '" + spec.getName().substring(0,30) + "'"));
            }
            if (!args.containsKey(Constants.PASS_URL_ARG)) {
                return errTextResp("Missing Argument", log.error("No Password Given"));
            }
            String message = "Player '" + spec.getName() + "' ";
            List<PlayerSpec> matches = players.filter(spec.samePrefixPred());
            if (matches.length() >= maxLogins) {
                return errTextResp("Multiple Registrations", log.error(message
                        + "Already Registered "+matches.length()+" time(s)"));
            }
            if (matches.contains(spec)){
                return errTextResp("Duplicate Registration", log.error(message
                        + "Already Registered at that Address"));
            }
            if (verify(passwds.get(spec.getName()), args.get(Constants.PASS_URL_ARG))) {
                if (spec.getAddress().equals(Constants.REG_AUTO)) {
                    spec = spec.changeAddress(sock.getInetAddress().getHostAddress());
                }
                PlayerSpec nspec = maxLogins > 1 ? spec.tagName(matches.length()): spec;
                players = players.push(nspec);
                saveFile();
                log.notify("Registering '" + spec.getName() + "' at " + spec.getAddress() + ":" + spec.getPort());
                log.event(message + "has been Registered");
                return okTextResp(spec.getName());
            }
            return errTextResp("Mismatch Password", log
                    .error("Password Does not Match Player '" + spec.getName() + "'"));
        } catch (ParseException pe) {
            return errTextResp("Parse Error", log.error("Error Parsing Registration (gen.PlayerSpec)"));
        }
    }

    /** Save the current Registration */
    public void saveFile(){
        if (savefile.length() == 0)
            return;
        log.notify("Saving Reg File: '" + savefile + "'");
        try {
            PrintStream outf = new PrintStream(new FileOutputStream(savefile));
            outf.print(players.toString("\n", ""));
            outf.close();
            log.notify("Done");
        } catch (IOException e) {
            log.error("Exception Saving Reg File: " + e);
        }
    }

    /** Encrypt and Verify the given Password agains the known 'Hash' */
    public static boolean verify(String known, String unknownPlain){
        return known.equals(Util.encrypt(unknownPlain));
    }

    /** Run the registration server, until the competition starts. */
    public List<PlayerSpec> runRegistration(String startDate) throws IOException, java.text.ParseException{
        return runRegistration(Option.some(Util.parseDate(startDate)));
    }

    /** Run the registration server, until the competition starts. */
    public List<PlayerSpec> runRegistration(Option<Date> start) throws IOException{
        ServerThread server = Factory.create(this,false,50);
        log.event("Registration Opened");
        if (start.isSome()) {
            log.notify("Competition will start at: " + Util.printDate(start.inner()));
            Util.waitUntil(start.inner());
        } else {
            System.err.print("Hit Enter to Close Reg: ");
            System.in.read();
        }
        server.shutdown();
        log.event("Registration Closed");
        return players;
    }

    /** Just the Registration */
    public static void main(String[] argStr) throws Exception{
        List<String> split[] = CLI.splitArgs(argStr), options = split[CLI.OPTS], args = split[CLI.ARGS];
        if (args.contains("--help") || args.isEmpty()) {
            System.err.println(" ** Usage: java reg.RegServer <Port> [--partial] [--savereg] [RegFile]");
            System.err.println(" Examples: java reg.RegServer 8000");
            System.err.println("           java reg.RegServer 8000 --savereg regfile.txt");
            return;
        }
        int port = Integer.parseInt(args.top());
        Logger log = Logger.text(System.err, Util.logFileName("reg"));
        RegServer server = RegServer.create(port, log,
                options.contains(admin.AdminMain.SAVE_PLAYERS) ? (args.length() > 1 ? args.lookup(1)
                        : Constants.PLAYERS_FILE) : "", options.contains(admin.AdminMain.PARTIAL_REG), 1);
        List<PlayerSpec> players = server.runRegistration(Option.<Date> none());
        System.err.println("\n ******* DONE : " + Util.printDate(Util.now()) + " ************");
    }
}
